@@ -24,6 +24,8 @@
 
 #include "system.h"
 
+#include "die.h"
+#include "ignore-value.h"
 #include "op-scanner.h"
 #include "op-defs.h"
 #include "op-parser.h"
@@ -160,7 +162,7 @@ set_op_params (struct fieldop *op)
         op->params.bin_bucket_size = _params[0].f;
       /* TODO: in the future, accept offset as well? */
       if (_params_used>1)
-        error (EXIT_FAILURE, 0, _("too many parameters for operation %s"),
+        die (EXIT_FAILURE, 0, _("too many parameters for operation %s"),
                                     quote (get_field_operation_name (op->op)));
       return;
     }
@@ -171,17 +173,31 @@ set_op_params (struct fieldop *op)
       if (_params_used==1)
         op->params.strbin_bucket_size = _params[0].u;
       if (op->params.strbin_bucket_size==0)
-        error (EXIT_FAILURE, 0, _("strbin bucket size must not be zero"));
+        die (EXIT_FAILURE, 0, _("strbin bucket size must not be zero"));
       /* TODO: in the future, accept offset as well? */
       if (_params_used>1)
-        error (EXIT_FAILURE, 0, _("too many parameters for operation %s"),
+        die (EXIT_FAILURE, 0, _("too many parameters for operation %s"),
+                                    quote (get_field_operation_name (op->op)));
+      return;
+    }
+
+  if (op->op==OP_PERCENTILE)
+    {
+      op->params.percentile = 95; /* default percentile */
+      if (_params_used==1)
+        op->params.percentile = _params[0].u;
+      if (op->params.percentile==0 || op->params.percentile>100)
+        die (EXIT_FAILURE, 0, _("invalid percentile value %" PRIuMAX),
+               op->params.percentile);
+      if (_params_used>1)
+        die (EXIT_FAILURE, 0, _("too many parameters for operation %s"),
                                     quote (get_field_operation_name (op->op)));
       return;
     }
 
   /* All other operations do not take parameters */
   if (_params_used>0)
-    error (EXIT_FAILURE, 0, _("too many parameters for operation %s"),
+    die (EXIT_FAILURE, 0, _("too many parameters for operation %s"),
         quote (get_field_operation_name (op->op)));
 }
 
@@ -198,44 +214,45 @@ parse_simple_operation_column (struct parser_field_t /*OUTPUT*/ *p,
       p->name = xstrdup (scanner_identifier);
       break;
 
-    case TOK_WHITESPACE:
+    case TOK_WHITESPACE:                        /* LCOV_EXCL_LINE */
       internal_error ("whitespace");            /* LCOV_EXCL_LINE */
 
     case TOK_COMMA:
-      error (EXIT_FAILURE, 0, _("missing field for operation %s"),
-          quote (get_field_operation_name (fop)));
+      die (EXIT_FAILURE, 0, _("missing field for operation %s"),
+           quote (get_field_operation_name (fop)));
 
     case TOK_END:
       /* informative error message depends on the context: */
       if (in_range)
-        error (EXIT_FAILURE, 0, _("invalid field range for operation %s"),
-            quote (get_field_operation_name (fop)));
+        die (EXIT_FAILURE, 0, _("invalid field range for operation %s"),
+             quote (get_field_operation_name (fop)));
       if (in_pair)
-        error (EXIT_FAILURE, 0, _("invalid field pair for operation %s"),
-            quote (get_field_operation_name (fop)));
-      error (EXIT_FAILURE, 0, _("missing field for operation %s"),
-          quote (get_field_operation_name (fop)));
+        die (EXIT_FAILURE, 0, _("invalid field pair for operation %s"),
+             quote (get_field_operation_name (fop)));
+      die (EXIT_FAILURE, 0, _("missing field for operation %s"),
+           quote (get_field_operation_name (fop)));
 
     case TOK_DASH:
-      error (EXIT_FAILURE, 0, _("invalid field range for operation %s"),
-          quote (get_field_operation_name (fop)));
+      die (EXIT_FAILURE, 0, _("invalid field range for operation %s"),
+           quote (get_field_operation_name (fop)));
 
     case TOK_COLONS:
-      error (EXIT_FAILURE, 0, _("invalid field pair for operation %s"),
-          quote (get_field_operation_name (fop)));
+      die (EXIT_FAILURE, 0, _("invalid field pair for operation %s"),
+           quote (get_field_operation_name (fop)));
 
     case TOK_INTEGER:
-      /* Zero values will fall-thought to the error message below */
+      /* Zero values will fall-through to the error message below */
       if (scan_val_int>0)
         {
           p->by_name = false;
           p->num = scan_val_int;
           break;
         }
+      /* fallthrough */
 
     case TOK_FLOAT:
     default:
-      error (EXIT_FAILURE, 0, _("invalid field '%s' for operation %s"),
+      die (EXIT_FAILURE, 0, _("invalid field '%s' for operation %s"),
           scanner_identifier,
           quote (get_field_operation_name (fop)));
     }
@@ -269,10 +286,10 @@ parse_operation_column ()
       parse_simple_operation_column (q, true, false);
 
       if (p->by_name || q->by_name)
-        error (EXIT_FAILURE, 0, _("field range for %s must be numeric"),
+        die (EXIT_FAILURE, 0, _("field range for %s must be numeric"),
                                 quote (get_field_operation_name (fop)));
       if (p->num >= q->num)
-        error (EXIT_FAILURE, 0, _("invalid field range for operation %s"),
+        die (EXIT_FAILURE, 0, _("invalid field range for operation %s"),
                                 quote (get_field_operation_name (fop)));
     }
 }
@@ -329,7 +346,7 @@ parse_operation_params ()
 
         case TOK_WHITESPACE:
         case TOK_END:
-          error (EXIT_FAILURE, 0, _("missing parameter for operation %s"),
+          die (EXIT_FAILURE, 0, _("missing parameter for operation %s"),
                                   quote (get_field_operation_name (fop)));
 
         case TOK_COMMA:
@@ -337,7 +354,7 @@ parse_operation_params ()
         case TOK_IDENTIFIER:
         case TOK_COLONS:
         default:
-          error (EXIT_FAILURE, 0, _("invalid parameter %s for operation %s"),
+          die (EXIT_FAILURE, 0, _("invalid parameter %s for operation %s"),
                                   scanner_identifier,
                                   quote (get_field_operation_name (fop)));
 
@@ -369,10 +386,10 @@ create_field_ops ()
       set_op_params (op);
 
       if (OP_NEED_PAIR_PARAMS (fop) && !f->pair)
-        error (EXIT_FAILURE, 0, _("operation %s requires field pairs"),
+        die (EXIT_FAILURE, 0, _("operation %s requires field pairs"),
                                 quote (get_field_operation_name (fop)));
       if (!OP_NEED_PAIR_PARAMS (fop) && f->pair)
-        error (EXIT_FAILURE, 0, _("operation %s cannot use pair of fields"),
+        die (EXIT_FAILURE, 0, _("operation %s cannot use pair of fields"),
                                 quote (get_field_operation_name (fop)));
 
       if (f->range)
@@ -412,15 +429,15 @@ parse_operation (enum processing_mode pm)
     {
       pm2 = get_processing_mode (scanner_identifier);
       if (pm2 != MODE_INVALID)
-        error (EXIT_FAILURE,0, _("conflicting operation %s"),
+        die (EXIT_FAILURE,0, _("conflicting operation %s"),
                                quote (scanner_identifier));
 
-      error (EXIT_FAILURE,0, _("invalid operation %s"),
+      die (EXIT_FAILURE,0, _("invalid operation %s"),
                              quote (scanner_identifier));
     }
 
   if (!compatible_operation_modes (pm,pm2))
-    error (EXIT_FAILURE, 0, _("conflicting operation found: "\
+    die (EXIT_FAILURE, 0, _("conflicting operation found: "\
            "expecting %s operations, but found %s operation %s"),
            get_processing_mode_name (pm),
            get_processing_mode_name (pm2),
@@ -465,12 +482,13 @@ parse_mode_column (enum processing_mode pm)
       ADD_NAMED_GROUP (scanner_identifier);
       break;
 
-    case TOK_WHITESPACE:
+    case TOK_WHITESPACE:                        /* LCOV_EXCL_LINE */
       internal_error ("whitespace");            /* LCOV_EXCL_LINE */
+
     case TOK_COMMA:
     case TOK_END:
-      error (EXIT_FAILURE, 0, _("missing field for operation %s"),
-          quote (get_processing_mode_name (pm)));
+      die (EXIT_FAILURE, 0, _("missing field for operation %s"),
+           quote (get_processing_mode_name (pm)));
 
     case TOK_INTEGER:
       if (scan_val_int>0)
@@ -478,12 +496,13 @@ parse_mode_column (enum processing_mode pm)
           ADD_NUMERIC_GROUP (scan_val_int);
           break;
         }
+      /* fallthrough */
 
     case TOK_DASH:
     case TOK_COLONS:
     case TOK_FLOAT:
     default:
-      error (EXIT_FAILURE, 0, _("invalid field '%s' for operation %s"),
+      die (EXIT_FAILURE, 0, _("invalid field '%s' for operation %s"),
           scanner_identifier,
           quote (get_processing_mode_name (pm)));
 
@@ -505,13 +524,82 @@ parse_mode_column_list (enum processing_mode pm)
   /* detect and warn about incorrect usage,
      field specification for groups can't handle dashes or colons (for now) */
   if (tok == TOK_DASH)
-    error (EXIT_FAILURE, 0, _("invalid field range for operation %s"),
+    die (EXIT_FAILURE, 0, _("invalid field range for operation %s"),
                             quote (get_processing_mode_name (pm)));
   if (tok == TOK_COLONS)
-    error (EXIT_FAILURE, 0, _("invalid field pair for operation %s"),
+    die (EXIT_FAILURE, 0, _("invalid field pair for operation %s"),
                             quote (get_processing_mode_name (pm)));
 }
 
+static bool
+parse_check_line_or_field (const char* s)
+{
+  if (STREQ (s,"lines") || STREQ (s,"line") \
+      || STREQ (s,"rows") || STREQ (s,"row"))
+    return true;
+  if (STREQ (s,"fields") || STREQ (s,"field") \
+      || STREQ (s,"columns") || STREQ (s,"column") || STREQ (s,"col"))
+    return false;
+
+  die (EXIT_FAILURE, 0, _("invalid option %s for operation check"), quote (s));
+}
+
+static void
+parse_mode_check ()
+{
+  bool set_lines = true; // false = set columns
+  uintmax_t value = 0;
+
+  uintmax_t n_lines = 0;
+  uintmax_t n_fields = 0;
+
+  enum TOKEN tok = scanner_peek_token ();
+  while (tok != TOK_END)
+    {
+      tok = scanner_get_token ();
+      if (tok == TOK_INTEGER)
+        {
+          value = scan_val_int;
+
+          ignore_value (scanner_get_token ());
+          set_lines = parse_check_line_or_field (scanner_identifier);
+        }
+      else
+        {
+          set_lines = parse_check_line_or_field (scanner_identifier);
+          tok = scanner_get_token ();
+          if (tok != TOK_INTEGER)
+            die (EXIT_FAILURE, 0, _("number expected after option in " \
+                                    "operation 'check'"));
+          value = scan_val_int;
+        }
+
+      if (value == 0)
+        die (EXIT_FAILURE, 0, _("invalid value zero for lines/fields in "  \
+                                "operation 'check'"));
+
+      if (set_lines)
+        {
+          if (n_lines>0)
+            die (EXIT_FAILURE, 0, _("number of lines/rows already set in " \
+                                    "operation 'check'"));
+          n_lines = value;
+        }
+      else
+        {
+          if (n_fields>0)
+            die (EXIT_FAILURE, 0, _("number of fields/columns already set in " \
+                                    "operation 'check'"));
+          n_fields = value;
+        }
+
+      tok = scanner_peek_token ();
+    }
+
+  dm->mode_params.check_params.n_lines = n_lines;
+  dm->mode_params.check_params.n_fields = n_fields;
+}
+
 static void
 parse_mode ()
 {
@@ -524,7 +612,10 @@ parse_mode ()
   case MODE_TRANSPOSE:
   case MODE_NOOP:
   case MODE_REVERSE:
+    break;
+
   case MODE_TABULAR_CHECK:
+    parse_mode_check ();
     break;
 
   case MODE_REMOVE_DUPS:
@@ -534,7 +625,7 @@ parse_mode ()
   case MODE_CROSSTAB:
     parse_mode_column_list (pm);
     if (dm->num_grps!=2)
-      error (EXIT_FAILURE,0, _("crosstab requires exactly 2 fields, " \
+      die (EXIT_FAILURE,0, _("crosstab requires exactly 2 fields, " \
                                "found %"PRIuMAX), (uintmax_t)dm->num_grps);
 
     /* if the user didn't specify an operation, print counts */
@@ -547,7 +638,7 @@ parse_mode ()
       }
     else if (dm->num_ops>1)
       {
-        error (EXIT_FAILURE,0, _("crosstab supports one operation, " \
+        die (EXIT_FAILURE,0, _("crosstab supports one operation, " \
                                  "found %"PRIuMAX), (uintmax_t)dm->num_ops);
       }
     break;
@@ -556,21 +647,21 @@ parse_mode ()
     parse_mode_column_list (pm);
     parse_operations (pm);
     if (dm->num_ops==0)
-      error (EXIT_FAILURE,0, _("missing operation"));
+      die (EXIT_FAILURE,0, _("missing operation"));
     break;
 
-  case MODE_PER_LINE:
+  case MODE_PER_LINE:                           /* LCOV_EXCL_LINE */
     internal_error ("line mode used directly"); /* LCOV_EXCL_LINE */
     break;
 
-  case MODE_INVALID:
-  default:
+  case MODE_INVALID:                 /* LCOV_EXCL_LINE */
+  default:                           /* LCOV_EXCL_LINE */
     internal_error ("wrong opmode"); /* LCOV_EXCL_LINE */
     break;
   }
 
   if (scanner_peek_token ()!=TOK_END)
-    error (EXIT_FAILURE,0,_("extra operand %s"), quote (scanner_identifier));
+    die (EXIT_FAILURE,0,_("extra operand %s"), quote (scanner_identifier));
 }
 
 static void
@@ -594,7 +685,7 @@ parse_mode_or_op ()
       return ;
     }
 
-  error (EXIT_FAILURE,0, _("invalid operation %s"),
+  die (EXIT_FAILURE,0, _("invalid operation %s"),
 		  quote (scanner_identifier));
 }
 
@@ -707,7 +798,7 @@ datamash_ops_debug_print ( const struct datamash_ops* p )
 int TESTMAIN (int argc, const char* argv[])
 {
   if (argc<2)
-    error (EXIT_FAILURE, 0, _("missing script (among arguments)"));
+    die (EXIT_FAILURE, 0, _("missing script (among arguments)"));
 
   struct datamash_ops *o = datamash_ops_parse (argc-1, argv+1);
   datamash_ops_debug_print ( o );
